pragma DEFJ(TIM1_INIT=0xFD)		//Timer 1 jako zegar UART (19200,n,8,1)
												//warto TH1 = 0xFD dla 11,0592 MHz
#pragma SMALL								//wybr modelu pamici
#include <stdio.h>						//funkcje wejcia-wyjcia
#include <string.h>					//operacje na acuchach tekstowych
#include <reg8252.h>					//definicje rejestrw AT89S8252
#include <eeprom8252.h>				//funkcje obsugi pamici EEPROM
#include <wdt8252.h>					//obsuga watchdoga
#define PACKETSIZE			128		//dugo pakietu przesyanych danych
#define NUMOFRPT				10			//liczba prb transmisji w przypadku 
												//bdu
#define LEDOFF				1			//definicje stanu diody LED
#define LEDON					0			

//kody bdw transmisji
#define NOCARRIER			0x01		//bd zestawienia poczenia
#define ERRORDURINGSEND	0x02		//bd podczas wysyania danych
#define ERRORDURINGREC	0x03		//bd podczas odbioru danych
#define GSMDISCONNECTED	0x04		//aparat GSM odczony
#define ERRORDURINGDIAL	0x05		//bd podczas wybierania numeru
#define ACK					0x07		//potwierdzenie
#define NACK					0x08		//brak potwierdzenia

//kody stanw programu
#define IDLE					0x00		//stan spoczynkowy
#define SEND					0x01		//wysyanie danych
#define RECEIVE				0x02		//odbieranie danych
#define DIAL					0x03		//wybieranie numeru telefonu
#define CHECK					0x04		//test poczenia z aparatem
#define STATUSADDR			0x01		//adres, pod ktrym zapamitany jest 
												//status w pamici EEPROM

//port doczenia diody LED
sbit LEDDIODE = P1^7;
//zmienna zawierajca stan programu
data BYTE STATUS;

//bufor danych do wysyki (1kB)
at 0x0000 xdata BYTE UDBUF[0x0400];
//i wskanik do bufora
xdata BYTE * tUDB;
//zmienna pomocnicza dla operacji na tekstach
idata BYTE stemp[10];
//timeout dla transmisji (jednostka to 1/14Hz)
data WORD TRTIMEOUT;


//funkcja przeprowadza reset software'owy
void softreset()
{
	IE = 0;									//mov IE,#0
	asm{0xD0, 0xE0};						//pop ACC
	asm{0xD0, 0xE0};						//pop ACC
	ACC = 0;									//mov A,#0
	asm{0xC0, 0xE0};						//push ACC
	asm{0xC0, 0xE0};						//push ACC
	asm{0x32}								//reti
}

//procedura obsugi przerwania od Timera 2
//odmierzanie czasu timeout
void TMR2_Isr(void) interrupt 6 using 1
{
	TF2 = 0;
	if (--TRTIMEOUT == 0) softreset();
}

//ustawienie limitu czasu operacji i uruchomienie przerwania Timera 2
void settimeout(WORD time)
{
	TRTIMEOUT = time;
	EA = TR2 = 1;
}

//ustawienie zmiennej status i zapis do EEPROM
void setstatus(BYTE x)
{
	STATUS = x;
	ee_savebyte(STATUS, STATUSADDR);
}

//funkcja realizuje opnienie k*1 ms dla rezonatora f=11,0592 MHz
void delay(WORD k)
{
	WORD i,j;
	for ( j = 0; j < k; ++j)
		for (i = 0; i <= 451; ++i);
}

/*wizualizacja komunikatw o bdach
 errormessage(A, B, C, D), gdzie:
 -A: liczba byskw LED
 -B: czas trwania zawiecenia diody LED
 -C: liczba powtrze
 -D: czas pomidzy powtrzeniami */
void errormessage(BYTE rpt, WORD ontime, BYTE loop, WORD delaybtw)
{
	BYTE i;

	while (loop > 0)
	{
		for (i = 0; i<rpt; i++)
		{
			LEDDIODE = LEDON;
			delay(ontime);
			LEDDIODE = LEDOFF;
			delay(350);
		}
		delay(delaybtw);
		loop--;
	}
}

//komunikaty o bdach (tu za pomoc diody LED)
void error(BYTE errcode)
{
	switch (errcode)
	{
		case NOCARRIER:
			errormessage(3, 150, 3, 2000);
			break;
		case ERRORDURINGSEND:
			errormessage(4, 150, 3, 2000);
			break;
		case ERRORDURINGREC:
			errormessage(5, 150, 3, 2000);
			break;
		case GSMDISCONNECTED:
			errormessage(6, 150, 3, 2000);
			break;
		case ERRORDURINGDIAL:
			errormessage(7, 100, 3, 2000);
			break;
		default:
			errormessage(1, 500, 3, 1000);
			break;
	}
}

//wysyanie danych, CRC obliczane podczas wysyki i doczane na kocu
//kadej paczki danych; przesyane pakiety maj dugo 128 bajtw
bit send(generic BYTE *buf, BYTE len)
{
	BYTE temp, crc = 0, i;

	wdtstart();							//uruchomienie watchdoga (timeout 
												//transmisji)
	wdtreset();
	for(i = 0; i<len; i++)			//przesanie len bajtw + bajt CRC
	{
		putchar(*buf);					//wysanie znaku przez UART
		crc ^= ~(*buf);					//wyliczenie CRC
		buf++;								//nastpna pozycja w buforze
		wdtreset();						//zerowanie watchdoga
	}
	temp = _getkey();					//pobranie CRC z odbiornika
	wdtstop();								//zatrzymanie watchdoga

	if (temp != crc) 					//jeli odebrana suma <> wyliczonej
	{
		putchar(NACK);					//wylij znak "brak potwierdzenia"
		return(0); 						//i zwr warto false
	}
		else
	{
		putchar(ACK);						//inaczej wylij potwierdzenie
		return(1);							//i zwr true
	}
}


//odbieranie danych, CRC obliczane w trakcie odbioru i sprawdzane
//po odebraniu paczki danych; funkcja zwraca true, jeli odebrane
//i wyliczone crc s ze sob zgodne
bit receive(generic BYTE *buf, BYTE len)
{
	BYTE crc = 0, temp, i;

	wdtstart();							//uruchomienie watchdoga (timeout 
												//transmisji)
	wdtreset();
	for (i = 0; i<len; i++)			//ptla dla len bajtw + bajt CRC
	{
		*buf = _getkey();				//pobranie znaku do bufora w RAM
		crc ^= ~(*buf);					//wyliczenie sumy kontrolnej
		buf++;								//nastpna pozycja w buforze
		wdtreset();						//zerowanie watchdoga
	}
	putchar(crc);							//wysanie CRC do nadajnika
	wdtreset();
	temp = _getkey();					//oczekiwanie na potwierdzenie
	wdtstop();								//zatrzymanie watchdoga
	if (temp != ACK) return(0); else return(1);
}


//funkcja wysya zawarto bufora UDBUF (1 kB)
bit sendbuffer()
{
	BYTE i, rpt;

	setstatus(SEND);
	tUDB = &UDBUF;						//wskazanie na pocztek bufora
	for (i = 0; i<8; i++)				//przelij 8 pakietw po 128 bajtw
	{
		rpt = NUMOFRPT;					//liczba prb powtrze transmisji
		if (send(tUDB, PACKETSIZE))	//przesanie pakietu danych
		{
			tUDB = tUDB + PACKETSIZE;	//nastpna pozycja w buforze
		}
			else
		{
			while (--rpt > 0)			//by bd transmisji, powtrz
			{
				if (send(tUDB, PACKETSIZE)) 	//jeli bez bdu, opu 
															//ptl
				break;
			}
			if (rpt == 0) return(0);	//jeli nadal bd, opu funkcj, 
												//zwr 0
		}
	}
	return(1);								//pomylnie przesano bufor
}

//funkcja odbiera 8 blokw po 128 bajtw kady
bit receivebuffer()
{
	BYTE i, rpt;

	setstatus(RECEIVE);
	tUDB = &UDBUF;						//wskazanie na pocztek bufora
	for (i = 0; i<8; i++)				//odbierz 8 pakietw po 128 bajtw
	{
		rpt = NUMOFRPT;					//liczba prb powtrze transmisji
		if (receive(tUDB, PACKETSIZE))	//odbir pakietu
		{
			tUDB = tUDB + PACKETSIZE;		//jeli pomylny - nastpna poz.
														//w buforze
		}
			else
		{
			while (--rpt > 0)			//by bd odbioru, powtrz
			{
				if (receive(tUDB, PACKETSIZE))
				break;						//jeli tym razem bez bdu - opu 													//ptl
			}
			if (rpt == 0) return(0);	//jeli nadal bd, zwr 0
		}
	}
	return(1);								//pomylnie odebrano bufor
}


//rozczenie poczenia
void disconnect()
{
	printf("%s\n", "ATH");
}


//odebranie poczenia przychodzcego
void hookup()
{
	printf("%s\n", "ATA");
}


//inicjalizacja aparatu GSM, zestawienie poczenia danych
bit dialnumber(generic BYTE * PHONENUM)
{

	setstatus(DIAL);
	settimeout(60);
	printf("%s%s\n", "ATD ", PHONENUM);	//wybranie numeru telefonu
	gets(&stemp);							//komunikat zwracany przez aparat GSM
	EA = TR2 = 0;							//zatrzymanie przerwania Timera 2
	if (strncmp(stemp, "CONNECT", 7)!= 0)
												//jeli komunikat<>CONNECT, to bd
	return(0); else return(1);
}


//funkcja wysya komend ATE0 do telefonu GSM i sprawdza, czy dotar 	//komunikat OK
//jeli tak, zwracana jest warto TRUE; dodatkowo jest uruchamiany 	//watchdog
bit gsmconnected()
{
	setstatus(CHECK);
	settimeout(5);						//ustawienie limitu czasu na 5 sekund
	printf("%s\n","ATE0");			//wyczenie echa komendy
	gets(&stemp);
	gets(&stemp);							//pobranie "OK"+0A+0D
	EA = TR2 = 0;							//zatrzymanie odmierzania timeout
	if (strncmp(stemp, "OK", 2) == 0) return(1); else return(0);
}


//program gwny, przykad aplikacji do przesyania danych
void main()
{
	P3 &= 0xCF;
	P1 &= 0xEF;
	RCAP2L = RCAP2H = T2CON = 0;	//nastawy Timera 2
	ET2 = 1;									//zezwolenie na przyjmowanie przerwa 
												//od Timera 2
	STATUS = ee_getbyte(STATUSADDR);

	if (STATUS == 0xFF) STATUS = IDLE;//STATUS nie by ustawiany
	if (STATUS != IDLE)				//jeli status<>idle, to 																	//interpretacja
	{											//statusu, komunikat o bdzie 															//i przejcie do
												//trybu idle
		disconnect();						//rozczenie ewentualnego poczenia
		switch (STATUS)
		{
			case SEND:
				error(ERRORDURINGSEND);
				break;
			case RECEIVE:
				error(ERRORDURINGREC);
				break;
			case CHECK:
				error(GSMDISCONNECTED);
				break;
			case DIAL:
				error(ERRORDURINGDIAL);
				break;
		}
		goto idlemode;
	}

	wdtset(0xE2);							//watchdog: nastawa 2048 ms (timeout 
												//transmisji)
	PCON |= 0x80;							//ustawienie bitu prescalera zegara 
												//transmisji

	if (gsmconnected())
	{
		errormessage(1, 100, 1, 1000);	//sygnalizacja (zawiecenie LED)
		if (dialnumber("60330461"))	//jeli udao si nawiza poczenie
		{										//wysyana jest zawarto bufora
			if (!sendbuffer()) error(NOCARRIER);
			disconnect();					//rozczenie toru transmisji
		}
			else error(ERRORDURINGDIAL);
	}
		else error(GSMDISCONNECTED);
	
idlemode:
	setstatus(IDLE);
	LEDDIODE = LEDON;
	while(1);
}
